

class Shot(object):
    """
    This class represents on shot and contains shot properties such as start/end frame index of a shot, shot-id and
    video_name.
    """

    def __init__(self, sid, movie_name, start_pos, end_pos):
        """
        Constructor.

        :param sid: This parameter is used to specify a shot id.
        :param movie_name: this parameter is used to specify a movie_name.
        :param start_pos: This parameter is used to specify the start frame index of a shot.
        :param end_pos: This parameter is used to specify the end frame index of a shot.
        """
        #print("create instance of shot ...");
        self.sid = sid
        self.movie_name = movie_name
        self.start_pos = start_pos
        self.end_pos = end_pos

    def convert2String(self):
        """
        This method is used to convert all properties of a shot into a semicolon-separated string.
        :return:
        """
        tmp_str = str(self.movie_name) + ";" + str(self.sid) + ";" + str(self.start_pos) + ";" + str(self.end_pos)
        return tmp_str

    def printShotInfo(self):
        """
        This method is used to print all properties of a shot.
        """
        print("------------------------")
        print("shot id: " + str(self.sid))
        print("movie name: " + str(self.movie_name))
        print("start frame: " + str(self.start_pos))
        print("end frame: " + str(self.end_pos))



from vhh_sbd.utils import *
import yaml


class Configuration(object):
    """
    This class is needed to read the configuration parameters specified in the configuration.yaml file.
    The instance of the class is holding all parameters during runtime.

    .. note::
       e.g. ./config/config_vhh_test.yaml

        the yaml file is separated in multiple sections
        config['Development']
        config['PreProcessing']
        config['SbdCore']
        config['Evaluation']

        whereas each section should hold related and meaningful parameters.
    """

    def __init__(self, config_file: str):
        """
        Constructor

        :param config_file: [required] path to configuration file (e.g. PATH_TO/config.yaml)
                                       must be with extension ".yaml"
        """
        printCustom("create instance of configuration ... ", STDOUT_TYPE.INFO)

        self.config_file = config_file

        self.debug_flag = -1
        self.use_gpu = False

        self.resize_dim = None
        self.flag_convert2Gray = -1
        self.flag_crop = -1
        self.flag_downscale = -1
        self.opt_histogram_equ = None

        self.activate_candidate_selection = -1

        self.save_raw_results = -1
        self.path_postfix_raw_results = None
        self.path_prefix_raw_results = None
        self.path_raw_results = None

        self.save_final_results = -1
        self.path_prefix_final_results = None
        self.path_postfix_final_results = None
        self.path_final_results = None

        self.path_videos = None
        self.threshold = 0.0
        self.alpha = 0.0
        self.threshold_mode = None
        self.window_size = None
        self.backbone_cnn = None
        self.similarity_metric = None

        self.pretrained_model = None

        self.path_eval_results = None
        self.path_raw_results_eval = None
        self.save_eval_results = -1
        self.path_gt_data = None

    def loadConfig(self):
        """
        Method to load configurables from the specified configuration file
        """

        fp = open(self.config_file, 'r')
        config = yaml.load(fp, Loader=yaml.BaseLoader)

        developer_config = config['Development']
        pre_processing_config = config['PreProcessing']
        sbd_core_config = config['SbdCore']
        candidate_selection_config = config['CandidateSelection']
        evaluation_config = config['Evaluation']

        # developer_config section
        self.debug_flag = int(developer_config['DEBUG_FLAG'])

        if int(developer_config['USE_GPU']) == 1:
            self.use_gpu = True

        # pre-processing section
        self.resize_dim = (int(pre_processing_config['RESIZE_DIM'].split(',')[0]),
                           int(pre_processing_config['RESIZE_DIM'].split(',')[1]))
        self.flag_convert2Gray = int(pre_processing_config['CONVERT2GRAY'])
        self.flag_crop = int(pre_processing_config['CROP'])
        self.flag_downscale = int(pre_processing_config['DOWNSCALE'])
        self.opt_histogram_equ = pre_processing_config['HISTOGRAM_EQU']

        # sbd_core_config section
        self.activate_candidate_selection = int(sbd_core_config['CANDIDATE_SELECTION'])
        self.save_raw_results = int(sbd_core_config['SAVE_RAW_RESULTS'])
        self.path_postfix_raw_results = sbd_core_config['POSTFIX_RAW_RESULTS']
        self.path_prefix_raw_results = sbd_core_config['PREFIX_RAW_RESULTS']
        self.path_raw_results = sbd_core_config['PATH_RAW_RESULTS']

        self.save_final_results = int(sbd_core_config['SAVE_FINAL_RESULTS'])
        self.path_prefix_final_results = sbd_core_config['PREFIX_FINAL_RESULTS']
        self.path_postfix_final_results = sbd_core_config['POSTFIX_FINAL_RESULTS']
        self.path_final_results = sbd_core_config['PATH_FINAL_RESULTS']

        self.path_videos = sbd_core_config['PATH_VIDEOS']
        self.threshold_mode = sbd_core_config['THRESHOLD_MODE']
        self.window_size = int(sbd_core_config['WINDOW_SIZE'])
        self.threshold = float(sbd_core_config['THRESHOLD'])
        self.alpha = float(sbd_core_config['ALPHA'])
        self.backbone_cnn = sbd_core_config['BACKBONE_CNN']
        self.similarity_metric = sbd_core_config['SIMILARITY_METRIC']

        # candidate_selection_config
        self.pretrained_model = candidate_selection_config['PATH_PRETRAINED_MODEL']

        # evaluation section
        self.path_raw_results_eval = evaluation_config['PATH_RAW_RESULTS']
        self.path_eval_results = evaluation_config['PATH_EVAL_RESULTS']
        self.save_eval_results = int(evaluation_config['SAVE_EVAL_RESULTS'])
        self.path_gt_data = evaluation_config['PATH_GT_ANNOTATIONS']



import numpy as np
import csv
from time import gmtime, strftime
import json
from scipy.spatial import distance
from vhh_sbd.Video import Video
import matplotlib.pyplot as plt
from inspect import signature
from vhh_sbd.Configuration import Configuration
from vhh_sbd.SBD import SBD
from vhh_sbd.Shot import Shot
import os


class Evaluation(object):
    """
    This class is used to evaluate the implemented algorithm.
    """

    def __init__(self, config_file: str):
        """
        Constructor.

        :param config_file: [required] path to configuration file (e.g. PATH_TO/config.yaml)
                                       must be with extension ".yaml"
        """
        print("create instance of evaluation ...")
        self.config_instance = Configuration(config_file)
        self.config_instance.loadConfig()

        self.precision = 0
        self.recall = 0
        self.f1score = 0


    '''
    def loadRawResultsAsCsv(self, filepath):
        # save raw results to file
        fp = open(filepath, mode='r');
        lines = fp.readlines();
        fp.close();
        # print(lines)

        dist_l = [];
        for i in range(0, len(lines)):
            line = lines[i].replace('\n', '')
            line = line.replace('[', '')
            line = line.replace(']', '')
            line = line.replace('\'', '')
            line = line.replace(',', '.')
            line = line.replace(' ', '')
            line_split = line.split(';')
            # print(line_split)
            vidname = line_split[0];
            distance = float(line_split[1]);
            dist_l.append([vidname, distance]);

        dist_np = np.array(dist_l)
        #print(dist_np.shape)
        return dist_np
    '''

    def calculateSimilarityMetric(self, results_np: np.ndarray, threshold=4.5):
        """
        This method is used to calculate the similarity metrics based on the pre-calculated raw results.

        :param results_np: This parameter must hold a valid numpy array.
        :param threshold: This parameter holds a threshold. (default: 4.5)
        :return: This method returns a numpy array including the final shot boundaries.
        """
        shot_boundaries_l = []
        for i in range(0, len(results_np)):
            vid_name = results_np[i][0]
            start = results_np[i][1]
            end = results_np[i][2]
            distances_l = results_np[i][3]
            distances_np = np.array(distances_l).astype('float')

            if(self.config_instance.activate_candidate_selection == 0):
                # just take all frame positions over specified threshold

                THRESHOLD_MODE = self.config_instance.threshold_mode
                if (THRESHOLD_MODE == "adaptive"):
                    shot_l = []
                    thresholds = []
                    window_size = self.config_instance.window_size
                    alpha = threshold
                    for i in range(0, len(distances_np)):
                        if (i % window_size == 0):
                            th = np.mean(distances_np[i:i + window_size]) * alpha
                        thresholds.append(th)
                    thresholds = np.array(thresholds)
                    print(thresholds.shape)

                    for i in range(0, len(distances_np)):
                        if (distances_np[i] > thresholds[i]):
                            idx_curr = i + 1
                            idx_prev = i
                            shot_boundaries_l.append([vid_name, idx_prev, idx_curr])
                            # print("cut at: " + str(i) + " -> " + str(i+1))
                            # print(i)
                            # print(thresholds[i])
                            # print(distances_np[i])
                if (THRESHOLD_MODE == "fixed"):
                    idx_max = np.where(distances_np > threshold)[0]
                    print(idx_max)

                    if(len(idx_max) == 1) :
                        final_idx = idx_max #+ start
                        shot_boundaries_l.append([vid_name, final_idx, final_idx + 1])
                    elif(len(idx_max) > 1):
                        final_idx = idx_max #+ start
                        print(final_idx)
                        #exit()
                        for a in range(0, len(final_idx)):
                            shot_boundaries_l.append([vid_name, final_idx[a], final_idx[a] + 1])

            elif(self.config_instance.activate_candidate_selection == 1):
                # just take all frame positions over specified threshold
                idx_max = np.argmax(distances_np)
                final_idx = idx_max + start
                shot_boundaries_l.append([vid_name, final_idx, final_idx + 1])

            # print(final_idx)

            # cv2.imwrite("./test_result" + str(i) + "_1.png", self.vid_instance.getFrame(idx_max[i]))
            # cv2.imwrite("./test_result" + str(i) + "_2.png", self.vid_instance.getFrame(idx_max[i] + 1))

        shot_boundaries_np = np.array(shot_boundaries_l)
        #print(shot_boundaries_np)
        return shot_boundaries_np

    def convertShotBoundaries2Shots(self, shot_boundaries_np: np.ndarray, vid_instance):
        """
        This method converts a list with detected shot boundaries to the final shots.

        :param shot_boundaries_np: This parameter must hold a numpy array with all detected shot boundaries.
        :return: This method returns a numpy list with the final shots.
        """
        # convert results to shot instances

        shot_l = []

        vidname_curr = shot_boundaries_np[0][0]
        start_curr, stop_curr = shot_boundaries_np[0][1]

        print("shot boundaries list: ")
        
        if (start_curr == -1 and stop_curr == -1):
            print("no shots detected ... ")
            shot = Shot(len(shot_boundaries_np), vidname_curr, 0, int(vid_instance.number_of_frames) - 1)
            shot_l.append(shot)
            return shot_l

        shot_start = 0
        shot_end = start_curr
        shot = Shot(1, vidname_curr, shot_start, shot_end)
        shot_l.append(shot)

        for i in range(1, len(shot_boundaries_np)):
            #print(i)
            start_prev, stop_prev = shot_boundaries_np[i-1][1]
            start_curr, stop_curr = shot_boundaries_np[i][1]
            vidname_curr = shot_boundaries_np[i][0]

            shot_start = int(stop_prev)
            shot_end = int(start_curr)
            shot = Shot(i + 1, vidname_curr, shot_start, shot_end)
            shot_l.append(shot)

        vidname_curr = shot_boundaries_np[-1][0]
        start_curr, stop_curr = shot_boundaries_np[-1][1]
        shot_start = int(stop_curr)
        shot_end = int(vid_instance.number_of_frames) - 1
        shot = Shot(len(shot_boundaries_np) + 1, vidname_curr, shot_start, shot_end)
        shot_l.append(shot)

        if(self.config_instance.debug_flag == 1):
            # print shot infos
            for i in range(0, len(shot_l)):
                shot_l[i].printShotInfo()

        return shot_l

    def convertShots2ShotBoundaries(self, shot_l: list):
        
        shot_boundaries_l = []
        for i in range(1, len(shot_l)):
            start_curr, stop_curr = shot_l[i][1]
            video_name_curr = shot_l[i][0]

            start_prev, stop_prev = shot_l[i-1][1]
            video_name_prev = shot_l[i-1][0]

            shot_boundaries_l.append([video_name_curr, (stop_prev, start_curr)])
        
        shot_boundaries_np = np.array(shot_boundaries_l)
        return shot_boundaries_np

    def load_results(self, filename):
        fp = open(filename, 'r')
        lines = fp.readlines()
        fp.close()

        data_l = []
        for line in lines[1:]:
            line = line.replace('\n', '')
            line = line.replace('\\', '/')
            #print(line)
            line_split = line.split(';')
            #print(line_split)
            data_l.append([line_split[0].split('.')[0], (int(line_split[2]), int(line_split[3]))])
        #data_np = np.array(gt_annotation_list)
        return data_l

    def evaluation(self, result_np, vid_name):
        """
        This method is needed to evaluate the gathered results for a specified video.

        :param result_np: This parameter must hold a valid numpy array.
        :param vid_name: This parameter represents a video name.
        :return: This method returns the calculated TP, TN, FP and FN counters.
        """
        src_path = self.config_instance.path_videos
        gt_path = self.config_instance.path_gt_data

        if (self.config_instance.path_postfix_raw_results == 'csv'):
            vid_name = vid_name.replace('results_raw_', '')
            vid_name = vid_name.replace('.csv', '')
        elif (self.config_instance.path_postfix_raw_results == 'npy'):
            vid_name = vid_name.replace('results_raw_', '')
            vid_name = vid_name.replace('.npy', '')

        video_obj = Video()
        video_obj.load(src_path + "/" + str(vid_name) + ".m4v")
        video_obj.printVIDInfo()
        
        # load groundtruth labels
        gt_file_list = os.listdir(gt_path)
        gt_file_list.sort()
        idx = gt_file_list.index(str(vid_name.split('.')[0]) + ".csv")

        file = gt_file_list[idx]
        #print(file)
        gt_l = self.load_results(gt_path + file)
        #print(gt_np)
        gt_np = np.array(gt_l)
        #print(gt_np[:10])

        sb_gt_np = self.convertShots2ShotBoundaries(gt_l)
        #print(sb_gt_np[:10])
        
        # load pred
        pred_l = []
        for i in range(0, len(result_np)):
            vidname = result_np[i][0]
            start = int(result_np[i][1])
            stop = int(result_np[i][2])
            pred_l.append([vidname, (start, stop)])

        sb_pred_np = np.array(pred_l)
        #print(sb_pred_np[:10])
        
        '''
        print(vid_name)
        idx = np.where(vid_name == pred_np)[0]
        sb_pred_np = pred_np[idx]
        print("---------")
        print(sb_pred_np)
        #print(gt_np[:10])
        idx = np.where(vid_name == all_gt_data_np)[0]
        sb_gt_np = all_gt_data_np[idx]
        print("---------")
        print(sb_gt_np)
        '''

        # video-based predictions
        tp_cnt = 0
        fp_cnt = 0
        tn_cnt = 0
        fn_cnt = 0
        for j in range(0, int(video_obj.number_of_frames)):
            curr_pos = j
            prev_pos = j - 1

            search_tuple = (prev_pos, curr_pos)
            if(len(sb_pred_np) > 0 ):
                list_pred_tmp = np.squeeze(sb_pred_np[:, 1:]).tolist()
            else:
                list_pred_tmp = []
            list_gt_tmp = np.squeeze(sb_gt_np[:, 1:]).tolist()

            gt_flag = False
            pred_flag = False

            try:
                # found tuple in pred
                res_pred = list_pred_tmp.index(search_tuple)
                #print(list_pred_tmp.index(search_tuple))
                pred_flag = True
            except:
                res_pred = 0

            try:
                # found tuple in gt
                res_gt = list_gt_tmp.index(search_tuple)
                #print(list_gt_tmp.index(search_tuple))
                gt_flag = True
            except:
                res_gt = 0

            #print(str(gt_flag) + " == " + str(pred_flag))
            tp_cond = gt_flag and pred_flag # find tuple in pred && find tuple in gt --> true
            fn_cond = gt_flag and not pred_flag # not find tuple in pred && find tuple in gt --> true
            fp_cond = not gt_flag and pred_flag
            tn_cond = not gt_flag and not pred_flag

            if(tp_cond == True):
                tp_cnt = tp_cnt + 1
            if (fp_cond == True):
                fp_cnt = fp_cnt + 1
            if (tn_cond == True):
                tn_cnt = tn_cnt + 1
            if (fn_cond == True):
                fn_cnt = fn_cnt + 1

        #precision, recall, accuracy, f1_score = self.calculateEvalMetrics(tp_cnt, fp_cnt, tn_cnt, fn_cnt);
        #tmp_str = str(vid_name) + ";" + str(tp_cnt) + ";" + str(fp_cnt) + ";" + str(tn_cnt) + ";" + str(
        #    fn_cnt) + ";" + str(precision) + ";" + str(recall) + ";" + str(accuracy) + ";" + str(f1_score)
        #print(tmp_str)
        '''
        print("---------------------------")
        print("video-based results")
        print("TP: " + str(tp_cnt))
        print("FP: " + str(fp_cnt))
        print("TN: " + str(tn_cnt))
        print("FN: " + str(fn_cnt))
        
        print("precision: " + str(precision))
        print("recall: " + str(recall))
        print("accuracy: " + str(accuracy))
        print("f1_score: " + str(f1_score))
        '''

        return tp_cnt, fp_cnt, tn_cnt, fn_cnt

    def calculateMetrics(self, tp_cnt, fp_cnt, tn_cnt, fn_cnt):
        """
        This method is used to calculate the evaluation metrics precision, recall and f1score.

        :param tp_cnt: This parameter must hold a valid integer representing the tp counter.
        :param fp_cnt: This parameter must hold a valid integer representing the fp counter.
        :param tn_cnt: This parameter must hold a valid integer representing the tn counter.
        :param fn_cnt: This parameter must hold a valid integer representing the fn counter.
        :return: This method returns the scores for precision, recall, accuracy, f1_score, tp_rate and fp_rate.
        """
        # calculate precision, recall,  accuracy
        if(tp_cnt + fp_cnt != 0):
            precision = tp_cnt / (tp_cnt + fp_cnt)
        else:
            precision = 0

        if (tp_cnt + fn_cnt != 0):
            recall = tp_cnt / (tp_cnt + fn_cnt)
        else:
            recall = 0

        if ((tp_cnt + tn_cnt + fp_cnt + fn_cnt) != 0):
            accuracy = (tp_cnt + tn_cnt) / (tp_cnt + tn_cnt + fp_cnt + fn_cnt)
        else:
            accuracy = 0

        if ((precision + recall) != 0):
            f1_score = 2 * (precision * recall) / (precision + recall)
        else:
            f1_score = 0

        if ((tp_cnt + fn_cnt) != 0):
            tp_rate = tp_cnt / (tp_cnt + fn_cnt)
        else:
            tp_rate = 0

        if ((tn_cnt + fp_cnt) != 0):
            #Specificity = True Negatives / (True Negatives + False Positives)
            fp_rate = 1 - (tn_cnt / (tn_cnt + fp_cnt))
        else:
            fp_rate = 0;

        return precision, recall, accuracy, f1_score, tp_rate, fp_rate

    def calculateEvaluationMetrics(self):
        """
        This method is used to calculate the evaluation metrics.

        :return: This methods returns a numpy array including a list of the calculated metrics (precision, recall, ...).
        """
        #print(self.config_instance.path_postfix_raw_results)
        if (self.config_instance.path_postfix_raw_results == 'csv'):
            vid_name_list = os.listdir(str(self.config_instance.path_raw_results_eval))
            vid_name_list = [i for i in vid_name_list if i.endswith('.csv')]
        elif (self.config_instance.path_postfix_raw_results == 'npy'):
            vid_name_list = os.listdir(str(self.config_instance.path_raw_results_eval))
            vid_name_list = [i for i in vid_name_list if i.endswith('.npy')]
        print(vid_name_list)

        final_results = []
        fp_video_based = None

        if(self.config_instance.threshold_mode == 'adaptive'):
            thresholds_l = [1.8, 1.9, 2.0, 2.1, 2.2, 2.3, 2.4, 2.5, 2.6, 2.7, 2.8, 2.9, 3.0,
                            3.1, 3.2, 3.3, 3.4, 3.5, 3.55, 3.6, 3.65, 3.7, 3.75, 4.0, 4.5, 5.0, 5.5, 6.0,
                            7.0, 7.5, 8.5, 9.0, 9.5, 10.5, 11.0, 11.5, 12.0, 12.5, 13.0, 13.5, 14.0, 14.5,
                            15.0, 15.5, 16.0
                            ]
        elif(self.config_instance.threshold_mode == 'fixed'):
            thresholds_l = [1.0, 0.95, 0.90, 0.85, 0.80, 0.75, 0.70, 0.65, 0.60, 0.55,
                            0.50, 0.45, 0.40, 0.35, 0.30, 0.25, 0.20, 0.15, 0.10, 0.05, 0.0]
            #thresholds_l = [0.80, 0.50, 0.30]
        else:
            thresholds_l = []


        for t in thresholds_l:
            tp_sum = 0
            fp_sum = 0
            tn_sum = 0
            fn_sum = 0
            THRESHOLD = t

            if(int(self.config_instance.save_eval_results) == 1):
                fp_video_based = open(self.config_instance.path_eval_results + "/final_results_th-" + str(THRESHOLD) + ".csv", 'w')
                header = "vid_name;tp;fp;tn;fn;p;r;acc;f1_score;tp_rate;fp_rate"
                fp_video_based.write(header + "\n")

            results_l = []
            for vid_name in vid_name_list:
                if(self.config_instance.path_postfix_raw_results == 'csv'):
                    results_np = self.loadRawResultsFromCsv(self.config_instance.path_raw_results_eval + "/" + vid_name)
                elif (self.config_instance.path_postfix_raw_results == 'npy'):
                    results_np = self.loadRawResultsFromNumpy(self.config_instance.path_raw_results_eval + "/" + vid_name)

                # calculate similarity measures of consecutive frames and threshold them
                shot_boundaries_np = self.calculateSimilarityMetric(results_np, threshold=THRESHOLD)
                print(shot_boundaries_np)

                # calculate evaluation metrics
                tp, fp, tn, fn = self.evaluation(shot_boundaries_np, vid_name)
                p, r, acc, f1_score, tp_rate, fp_rate = self.calculateMetrics(tp, fp, tn, fn)

                if (int(self.config_instance.save_eval_results) == 1):
                    tmp_str = str(vid_name.replace('results_raw_', '').split('.')[0]) + ";" + str(tp) + ";" + str(fp) + \
                              ";" + str(tn) + ";" + str(fn) + ";" + str(p) + ";" + str(r) + ";" + str(acc) + ";" + \
                              str(f1_score) + ";" + str(tp_rate) + ";" + str(fp_rate)
                    print(tmp_str)
                    fp_video_based.write(tmp_str + "\n")
                #else:
                #    tp = 0;
                #    fp = 0;
                #    tn = 0;
                #    fn = 0;
                #    p = 0;
                #    r = 0;
                #    acc = 0;
                #    f1_score = 0;
                results_l.append([vid_name, tp, fp, tn, fn, p, r, acc, f1_score])

                tp_sum = tp_sum + tp
                fp_sum = fp_sum + fp
                tn_sum = tn_sum + tn
                fn_sum = fn_sum + fn

            p, r, acc, f1_score, tp_rate, fp_rate = self.calculateMetrics(tp_sum, fp_sum, tn_sum, fn_sum)

            if (int(self.config_instance.save_eval_results) == 1):
                tmp_str = str("overall" + ";" + str(tp_sum) + ";" + str(fp_sum) + \
                          ";" + str(tn_sum) + ";" + str(fn_sum) + ";" + str(p) + ";" + str(r) + ";" + str(acc) + ";" + \
                          str(f1_score) + ";" + str(tp_rate) + ";" + str(fp_rate))
                print(tmp_str)

                fp_video_based.write(tmp_str + "\n")
                fp_video_based.close()

            final_results.append([str(THRESHOLD), tp_sum, fp_sum, tn_sum, fn_sum, p, r, acc, f1_score, tp_rate, fp_rate])
        final_results_np = np.array(final_results)
        return final_results_np

    def plotPRCurve(self, results_np):
        """
        This method is needed to create and plot the precision_recall curve.

        :param results_np: This parameter must hold a vaild numpy array including the precision and recall scores.
        """
        print("plot precision recall curve ... ")
        precision = np.squeeze(results_np[1:, 5:6].astype('float')).tolist()
        recall = np.squeeze(results_np[1:, 6:7].astype('float')).tolist()

        plt.figure()
        step_kwargs = ({'step': 'post'}
                       if 'step' in signature(plt.fill_between).parameters
                       else {})
        plt.step(recall, precision, color='b', alpha=0.2, where='post')
        plt.fill_between(recall, precision, alpha=0.2, color='b', **step_kwargs)

        plt.xlabel('Recall')
        plt.ylabel('Precision')
        plt.ylim([0.0, 1.05])
        plt.xlim([0.0, 1.0])
        plt.title("2-class Precision-Recall curve");
        plt.savefig(self.config_instance.path_eval_results + "/pr_curve.png")

    def plotROCCurve(self, results_np):
        """
        This method is needed to create and plot the roc curve.

        :param results_np: This parameter must hold a vaild numpy array including the precision and recall scores.
        """

        print("plot roc curve ... ")
        tp_rate = np.squeeze(results_np[1:, 8:9].astype('float')).tolist()
        fp_rate = np.squeeze(results_np[1:, 9:10].astype('float')).tolist()

        plt.figure()
        plt.plot(fp_rate, tp_rate, color='orange', label='ROC')
        #plt.plot([0, 1], [0, 1], color='darkblue', linestyle='--')
        plt.xlabel('False Positive Rate')
        plt.ylabel('True Positive Rate')
        plt.title('Receiver Operating Characteristic (ROC) Curve')
        plt.legend()
        plt.savefig(self.config_instance.path_eval_results + "/roc_curve.png")

    def export2CSV(self, data_np: np.ndarray, header: str, filename: str, path: str):
        """
        This method is used to export the gathered results to a csv file.

        :param data_np: This parameter holds a valid numpy array.
        :param header: This parameter holds a csv header line (first line in the file - semicolon seperated).
        :param filename: This parameter must hold a valid file name.
        :param path: THis parameter must hold a valid file path.
        """
        # save to csv file
        fp = open(path + "/" + str(filename) + ".csv", 'w')
        fp.write(header)
        for i in range(0, len(data_np)):
            tmp_str = data_np[i][0]
            for j in range(1, len(data_np[0])):
                tmp_str = tmp_str + ";" + data_np[i][j]
                # print(str(i) + "/" + str(j) + " - " + tmp_str)
            fp.write(tmp_str + "\n")
        fp.close()

    def exportMovieResultsToCSV(self, fName, res_np):
        """
        This method is used to export video results to csv file.

        :param filepath: This parameter must hold a valid file_path.
        :param res_np: This parameter must hold a valid numpy array containing the final results.
        """
        print("start csv export ...")
        fName = fName.split('.')[0]
        print(fName)

        fp = open(str(fName) + "_movie_based.csv", mode='w')

        tmp_str = "vidname;precision;recall;f1score"
        fp.write(tmp_str + "\n")
        for i in range(0, len(res_np)):
            #print(sb_np[i])
            vidname = res_np[i][0]
            p = res_np[i][1]
            r = res_np[i][2]
            f1 = res_np[i][3]
            tmp_str = vidname + ";" + str(p) + ";" + str(r) + ";" + str(f1)
            fp.write(tmp_str + "\n")
            #csv_writer.writerow(row);

        fp.close()

    def loadResultsFromCSV(self, filepath):
        """
        This method is used to load final results from csv file.

        :param filepath: This parameter must hold a valid file_path.
        :return: This method returns a numpy array containing the final results.
        """

        print("load data from csv file ...")
        fp = open(filepath, mode='r')
        csv_reader = csv.reader(fp, delimiter=';')
        res_l = []

        for line in csv_reader:
            res_l.append(line)
        res_np = np.array(res_l)
        #print(res_np)

        fp.close()
        return res_np

    def loadRawResultsFromNumpy(self, filepath):
        """
        This method is used to load raw results from numpy array.

        :param filepath: This parameter must hold a valid file_path.
        :return: This method returns a numpy array containing the raw_results.
        """

        # save raw results to file
        print("load raw results from numpy ...")
        vid_name = filepath.split('/')[-1].split('.')[0]
        raw_results = np.load(filepath, allow_pickle=True)

        start = raw_results[0][0]
        stop = raw_results[0][1]
        dist_l = raw_results[0][2]
        dist_np = np.array(dist_l).astype('float')

        array_list = []
        array_list.append([vid_name.replace("results_raw_", ""), start, stop, dist_np])
        array_np = np.array(array_list)
        return array_np;

    def loadRawResultsFromCsv(self, filepath):
        """
        This method is used to load raw results from csv file.

        :param filepath: This parameter must hold a valid file_path.
        :return: This method returns a numpy array containing the raw_results.
        """
        # save raw results to file
        fp = open(filepath, mode='r')
        lines = fp.readlines()
        fp.close()
        #print(len(lines))

        final_l = []
        for i in range(0, len(lines)):
            line = lines[i].replace('\n', '')
            line = line.replace('[', '')
            line = line.replace(']', '')
            line = line.replace(')', '')
            line = line.replace('(', '')
            #line = line.replace('list', '')
            line = line.replace('\'', '')
            line = line.replace(',', '.')
            #line = line.replace(' ', '')
            line_split = line.split(';')
            vidname = line_split[0]
            start = int(line_split[1])
            end = int(line_split[2])
            dist_l = line_split[3:]
            dist_np = np.array(dist_l).astype('float')
            final_l.append([vidname, start, end, dist_np])

        final_np = np.array(final_l)
        return final_np

    def run(self):
        """
        This method is needed to run the evaluation process.
        """
        print("evaluation ... ")

        final_results_np = self.calculateEvaluationMetrics()
        print(final_results_np)
        self.plotPRCurve(final_results_np)
        self.plotROCCurve(final_results_np)

        '''
        # export results to csv file
        self.export2CSV(final_results_np,
                                     "threshold;p;r;acc;f1_score" + "\n",
                                     "final_evaluation_pr_curve",
                                     "../Develop/");
        '''


from vhh_sbd.utils import *
import numpy as np
import torch
from torchvision import models, transforms
from torch.autograd import Variable
from PIL import Image


class PyTorchModel(object):
    """
    This class is needed to create a specified cnn model architecture and extract the the features.
    """

    def __init__(self, model_arch, use_gpu: bool=False):
        """
        Constructor.
        :param model_arch: This parameter must hold a string containing a valid model architecture name.
        """
        printCustom("create instance of PyTorchModel ... ", STDOUT_TYPE.INFO)
        self.model_arch = model_arch
        if(self.model_arch == "squeezenet"):
            self.model = models.squeezenet1_0(pretrained=True)
        elif (self.model_arch == "vgg16"):
            self.model = models.vgg16(pretrained=True)
        else:
            self.model_arch = None
            printCustom("No valid backbone cnn network selected!", STDOUT_TYPE.ERROR)
            exit()

        self.normalize = transforms.Normalize(mean=[0.485, 0.456, 0.406], std=[0.229, 0.224, 0.225])

        self.use_gpu = use_gpu and torch.cuda.is_available()
        device = "GPU" if self.use_gpu else "CPU"
        printCustom(f"PyTorchModel created for inference on " + device, STDOUT_TYPE.INFO)

    def getFeatures(self, frm: np.ndarray):
        """
        This method is used to extract features.
        :param frm: THis parameter must hold a valid numpy image.
        :return: This method returns a flattened numpy array representing the visual features.
        """
        # print("calculate features ... ")
        try:
            image = Image.fromarray(frm.astype('uint8'))
            loader = transforms.Compose([transforms.ToTensor()])
            image = loader(image).float()
            image = self.normalize(image)

            image = Variable(image, requires_grad=True)
            image = image.unsqueeze(0)  # this is for VGG, may not be needed for ResNet

            # do inference on CPU or GPU depending on CUDA availability
            if self.use_gpu:
                self.model.features = self.model.features.cuda()
                inputs = image.cuda()
            else:
                self.model.features = self.model.features.cpu()
                inputs = image.cpu()

            self.model.features.eval()

            with torch.no_grad():
                outputs = self.model.features(inputs)
                outputs_flatten = outputs.view(outputs.size(0), -1)
                #print(outputs_flatten.size())
        except:
            print("+++++++++++++++++++++++++++++")
            exit(1)

        return outputs_flatten.cpu().detach().numpy()



import torch
import torch.nn as nn
import torch.nn.functional as F
import os
import cv2
from PIL import Image
import numpy as np
import random
import math
import numbers
import collections
from torch.autograd import Variable
try:
    import accimage
except ImportError:
    accimage = None
from vhh_sbd.Configuration import Configuration
from vhh_sbd.PreProcessing import PreProcessing


class deepSBD(nn.Module):
    """
    This class is represents the pytorch model architecture used for sbd candidate selection.
    It detects frames ranges of about 16 frames which includes an abrupt cut. The loaded model
    is pre-trained on the deepsbd dataset.
    """

    def __init__(self):
        """
        Constructor.
        """
        super(deepSBD, self).__init__()
        self.conv1 = nn.Conv3d(3, 96, kernel_size=3, stride=(1, 2, 2), padding=(0, 0, 0), bias=True)
        self.relu1 = nn.ReLU(inplace=True)
        self.pool1 = nn.MaxPool3d(kernel_size=(3, 3, 3), stride=(1, 2, 2), padding=0)
        self.conv2 = nn.Conv3d(96, 256, kernel_size=3, stride=(1, 2, 2), padding=(0, 0, 0), bias=True)
        self.relu2 = nn.ReLU(inplace=True)
        self.pool2 = nn.MaxPool3d(kernel_size=(3, 3, 3), stride=(1, 2, 2), padding=0)
        self.conv3 = nn.Conv3d(256, 384, kernel_size=3, stride=1, padding=1, bias=True)
        self.relu3 = nn.ReLU(inplace=True)
        self.conv4 = nn.Conv3d(384, 384, kernel_size=3, stride=1, padding=1, bias=True)
        self.relu4 = nn.ReLU(inplace=True)
        self.conv5 = nn.Conv3d(384, 256, kernel_size=3, stride=1, padding=1, bias=True)
        self.relu5 = nn.ReLU(inplace=True)
        self.pool1 = nn.MaxPool3d(kernel_size=(3, 3, 3), stride=(1, 2, 2), padding=0)
        self.fc6 = nn.Linear(100352, 2048)
        self.relu6 = nn.ReLU(inplace=True)
        self.fc7 = nn.Linear(2048, 2048)
        self.relu7 = nn.ReLU(inplace=True)
        self.fc8 = nn.Linear(2048, 3)
    
    def forward(self, x):
        """
        This method is needed to calculate the foward pass of the model.

        :param x: this parameter must be a valid pytorch tensor.
        :return:  This method returns a pytorch tensor with the specified shape.
        """
        x = self.conv1(x)
        x = self.relu1(x)
        x = self.pool1(x)
        x = self.conv2(x)
        x = self.relu2(x)
        x = self.pool2(x)
        x = self.conv3(x)
        x = self.relu3(x)
        x = self.conv4(x)
        x = self.relu4(x)
        x = self.conv5(x)
        x = self.relu5(x)
        x = x.view(x.size(0), -1)
        x = self.fc6(x)
        x = self.relu6(x)
        x = self.fc7(x)
        x = self.relu7(x)
        x = self.fc8(x)
        return x


class CandidateSelection(object):
    """
    This class is used for sbd candidate selection. It detects frames ranges of about 16 frames which includes an
    abrupt cut. The loaded model is pre-trained on the deepsbd dataset.
    """

    def __init__(self, config_instance: Configuration):
        """
        Constructor.

        :param config_instance: object instance of type Configuration
        """
        print("create instance of candidate selection module ... ")
        self.config_instance = config_instance;
        self.preprocessing = PreProcessing(self.config_instance)

    def run(self, video_path):
        """
        This method is used to run the candidate selection process.

        :param video_path: This parameter must hold a valid path to a video file.
        :return: This method returns a numpy array with a list of all detected frames ranges.
        """

        model_path = self.config_instance.pretrained_model
        temporal_length = 16
        batch_size = 32

        num_classes = 2
        self.model = deepSBD()  # deepsbd class
        self.model = self.model.cuda()
        self.model = nn.DataParallel(self.model)

        # /caa/Projects02/vhh/private/database_nobackup/public_datasets/Clipshots/ClipShots/
        # --model alexnet
        # --weights /caa/Homes01/dhelm/working/pycharm_sbd_clipshot/ClipShots_basline/pretrained/Alexnet-final.pth
        # --result_dir /caa/Homes01/dhelm/results/results_sbd/results_sbd_clipshots /
        # --test_list_path /caa/Projects02/vhh/private/database_nobackup/public_datasets/Clipshots/ClipShots/video_lists/test.txt

        checkpoint = torch.load(model_path)
        self.model.load_state_dict(checkpoint['state_dict'])

        self.spatial_transform = Compose([Scale((128, 128)),
                                          ToTensor(1),
                                          Normalize(self.get_mean(1), [1, 1, 1])])
        self.model.eval()
        videocap = cv2.VideoCapture(video_path)
        status = True
        clip_batch = []
        labels = []
        image_clip = []
        while status:
            for i in range(temporal_length - len(image_clip)):
                status, frame = videocap.read()
                #print(i)
                if not status:
                    break
                else:
                    # apply preprecessing
                    frame = cv2.cvtColor(frame, cv2.COLOR_BGR2RGB);
                    frame = self.preprocessing.applyTransformOnImg(frame)

                    frame = Image.fromarray(cv2.cvtColor(frame, cv2.COLOR_BGR2RGB)).convert('RGB')
                    frame = self.spatial_transform(frame)
                    image_clip.append(frame)

            image_clip += [image_clip[-1] for _ in range(temporal_length - len(image_clip))]

            if len(image_clip) == temporal_length:
                clip = torch.stack(image_clip, 0).permute(1, 0, 2, 3)
                clip_batch.append(clip)
                image_clip = image_clip[int(temporal_length / 2):]

            if len(clip_batch) == batch_size or not status:
                clip_tensor = torch.stack(clip_batch, 0)
                clip_tensor = Variable(clip_tensor).cuda()
                results = self.model(clip_tensor)
                labels += self.get_label(results)
                clip_batch = []

        final_res = []
        i = 0
        while i < len(labels):
            if labels[i] > 0:
                label = labels[i]
                begin = i
                i += 1
                while i < len(labels) and labels[i] == labels[i - 1]:
                    i += 1
                end = i - 1
                final_res.append((begin * temporal_length / 2 + 1, end * temporal_length / 2 + 16 + 1, label))
            else:
                i += 1

        # prepare results
        abrupt_l = []
        graduals_l = []
        for begin, end, label in final_res:
            if label == 2:
                abrupt_l.append((begin, end))
            else:
                graduals_l.append((begin, end))

        #print(final_res)
        #print(abrupt_l)
        #print(graduals_l)

        final_res_np = np.array(final_res)
        abrupt_np = np.array(abrupt_l).astype('int')
        graduals_np = np.array(graduals_l)
        return abrupt_np

    def get_mean(self, norm_value=255):
        """
        Helper method to calculate the normalized mean values.

        :param norm_value: Base for normalization (default: 255).
        :return: array with the normalized mean values for each color channel (RGB)
        """
        return [114.7748 / norm_value, 107.7354 / norm_value, 99.4750 / norm_value]

    def get_test_spatial_transform(self, opt):
        return Compose([Scale((opt.spatial_size, opt.spatial_size)),
                        ToTensor(opt.norm_value),
                        Normalize(self.get_mean(opt.norm_value), [1, 1, 1])])

    def get_label(self, res_tensor):
        res_numpy = res_tensor.data.cpu().numpy()
        labels = []
        for row in res_numpy:
            labels.append(np.argmax(row))
        return labels



class Compose(object):
    def __init__(self, transforms):
        self.transforms = transforms

    def __call__(self, img):
        for t in self.transforms:
            img = t(img)
        return img

    def randomize_parameters(self):
        for t in self.transforms:
            t.randomize_parameters()


class ToTensor(object):
    """Convert a ``PIL.Image`` or ``numpy.ndarray`` to tensor.
    Converts a PIL.Image or numpy.ndarray (H x W x C) in the range
    [0, 255] to a torch.FloatTensor of shape (C x H x W) in the range [0.0, 1.0].
    """

    def __init__(self, norm_value=255):
        self.norm_value = norm_value

    def __call__(self, pic):
        """
        Args:
            pic (PIL.Image or numpy.ndarray): Image to be converted to tensor.
        Returns:
            Tensor: Converted image.
        """
        if isinstance(pic, np.ndarray):
            # handle numpy array
            img = torch.from_numpy(pic.transpose((2, 0, 1)))
            # backward compatibility
            return img.float().div(self.norm_value)

        if accimage is not None and isinstance(pic, accimage.Image):
            nppic = np.zeros([pic.channels, pic.height, pic.width], dtype=np.float32)
            pic.copyto(nppic)
            return torch.from_numpy(nppic)

        # handle PIL Image
        if pic.mode == 'I':
            img = torch.from_numpy(np.array(pic, np.int32, copy=False))
        elif pic.mode == 'I;16':
            img = torch.from_numpy(np.array(pic, np.int16, copy=False))
        else:
            img = torch.ByteTensor(torch.ByteStorage.from_buffer(pic.tobytes()))
        # PIL image mode: 1, L, P, I, F, RGB, YCbCr, RGBA, CMYK
        if pic.mode == 'YCbCr':
            nchannel = 3
        elif pic.mode == 'I;16':
            nchannel = 1
        else:
            nchannel = len(pic.mode)
        img = img.view(pic.size[1], pic.size[0], nchannel)
        # put it from HWC to CHW format
        # yikes, this transpose takes 80% of the loading time/CPU
        img = img.transpose(0, 1).transpose(0, 2).contiguous()
        if isinstance(img, torch.ByteTensor):
            return img.float().div(self.norm_value)
        else:
            return img

    def randomize_parameters(self):
        pass


class Normalize(object):
    """Normalize an tensor image with mean and standard deviation.
    Given mean: (R, G, B) and std: (R, G, B),
    will normalize each channel of the torch.*Tensor, i.e.
    channel = (channel - mean) / std
    Args:
        mean (sequence): Sequence of means for R, G, B channels respecitvely.
        std (sequence): Sequence of standard deviations for R, G, B channels
            respecitvely.
    """

    def __init__(self, mean, std):
        self.mean = mean
        self.std = std

    def __call__(self, tensor):
        """
        Args:
            tensor (Tensor): Tensor image of size (C, H, W) to be normalized.
        Returns:
            Tensor: Normalized image.
        """
        # TODO: make efficient
        for t, m, s in zip(tensor, self.mean, self.std):
            t.sub_(m).div_(s)
        return tensor

    def randomize_parameters(self):
        pass


class Scale(object):
    """Rescale the input PIL.Image to the given size.
    Args:
        size (sequence or int): Desired output size. If size is a sequence like
            (w, h), output size will be matched to this. If size is an int,
            smaller edge of the image will be matched to this number.
            i.e, if height > width, then image will be rescaled to
            (size * height / width, size)
        interpolation (int, optional): Desired interpolation. Default is
            ``PIL.Image.BILINEAR``
    """

    def __init__(self, size, interpolation=Image.BILINEAR):
        assert isinstance(size, int) or (isinstance(size, collections.Iterable) and len(size) == 2)
        self.size = size
        self.interpolation = interpolation

    def __call__(self, img):
        """
        Args:
            img (PIL.Image): Image to be scaled.
        Returns:
            PIL.Image: Rescaled image.
        """
        if isinstance(self.size, int):
            w, h = img.size
            if (w <= h and w == self.size) or (h <= w and h == self.size):
                return img
            if w < h:
                ow = self.size
                oh = int(self.size * h / w)
                return img.resize((ow, oh), self.interpolation)
            else:
                oh = self.size
                ow = int(self.size * w / h)
                return img.resize((ow, oh), self.interpolation)
        else:
            return img.resize(self.size, self.interpolation)

    def randomize_parameters(self):
        pass


class CenterCrop(object):
    """Crops the given PIL.Image at the center.
    Args:
        size (sequence or int): Desired output size of the crop. If size is an
            int instead of sequence like (h, w), a square crop (size, size) is
            made.
    """

    def __init__(self, size):
        if isinstance(size, numbers.Number):
            self.size = (int(size), int(size))
        else:
            self.size = size

    def __call__(self, img):
        """
        Args:
            img (PIL.Image): Image to be cropped.
        Returns:
            PIL.Image: Cropped image.
        """
        w, h = img.size
        th, tw = self.size
        x1 = int(round((w - tw) / 2.))
        y1 = int(round((h - th) / 2.))
        return img.crop((x1, y1, x1 + tw, y1 + th))

    def randomize_parameters(self):
        pass


class CornerCrop(object):
    def __init__(self, size, crop_position=None):
        self.size = size
        if crop_position is None:
            self.randomize = True
        else:
            self.randomize = False
        self.crop_position = crop_position
        self.crop_positions = ['c', 'tl', 'tr', 'bl', 'br']

    def __call__(self, img):
        image_width = img.size[0]
        image_height = img.size[1]

        if self.crop_position == 'c':
            th, tw = (self.size, self.size)
            x1 = int(round((image_width - tw) / 2.))
            y1 = int(round((image_height - th) / 2.))
            x2 = x1 + tw
            y2 = y1 + th
        elif self.crop_position == 'tl':
            x1 = 0
            y1 = 0
            x2 = self.size
            y2 = self.size
        elif self.crop_position == 'tr':
            x1 = image_width - self.size
            y1 = 0
            x2 = image_width
            y2 = self.size
        elif self.crop_position == 'bl':
            x1 = 0
            y1 = image_height - self.size
            x2 = self.size
            y2 = image_height
        elif self.crop_position == 'br':
            x1 = image_width - self.size
            y1 = image_height - self.size
            x2 = image_width
            y2 = image_height

        img = img.crop((x1, y1, x2, y2))

        return img

    def randomize_parameters(self):
        if self.randomize:
            self.crop_position = self.crop_positions[
                random.randint(0, len(self.crop_positions) - 1)]


class RandomHorizontalFlip(object):
    """Horizontally flip the given PIL.Image randomly with a probability of 0.5."""

    def __call__(self, img):
        """
        Args:
            img (PIL.Image): Image to be flipped.
        Returns:
            PIL.Image: Randomly flipped image.
        """
        if self.p < 0.5:
            return img.transpose(Image.FLIP_LEFT_RIGHT)
        return img

    def randomize_parameters(self):
        self.p = random.random()


class MultiScaleCornerCrop(object):
    """Crop the given PIL.Image to randomly selected size.
    A crop of size is selected from scales of the original size.
    A position of cropping is randomly selected from 4 corners and 1 center.
    This crop is finally resized to given size.
    Args:
        scales: cropping scales of the original size
        size: size of the smaller edge
        interpolation: Default: PIL.Image.BILINEAR
    """

    def __init__(self, scales, size, interpolation=Image.BILINEAR):
        self.scales = scales
        self.size = size
        self.interpolation = interpolation

        self.crop_positions = ['c', 'tl', 'tr', 'bl', 'br']

    def __call__(self, img):
        min_length = min(img.size[0], img.size[1])
        crop_size = int(min_length * self.scale)

        image_width = img.size[0]
        image_height = img.size[1]

        if self.crop_position == 'c':
            center_x = image_width // 2
            center_y = image_height // 2
            box_half = crop_size // 2
            x1 = center_x - box_half
            y1 = center_y - box_half
            x2 = center_x + box_half
            y2 = center_y + box_half
        elif self.crop_position == 'tl':
            x1 = 0
            y1 = 0
            x2 = crop_size
            y2 = crop_size
        elif self.crop_position == 'tr':
            x1 = image_width - crop_size
            y1 = 0
            x2 = image_width
            y2 = crop_size
        elif self.crop_position == 'bl':
            x1 = 0
            y1 = image_height - crop_size
            x2 = crop_size
            y2 = image_height
        elif self.crop_position == 'br':
            x1 = image_width - crop_size
            y1 = image_height - crop_size
            x2 = image_width
            y2 = image_height

        img = img.crop((x1, y1, x2, y2))

        return img.resize((self.size, self.size), self.interpolation)

    def randomize_parameters(self):
        self.scale = self.scales[random.randint(0, len(self.scales) - 1)]
        self.crop_position = self.crop_positions[random.randint(0, len(self.scales) - 1)]




'''

# spatial_transforms = get_test_spatial_transform(opt)
spatial_transforms = Compose([Scale((128, 128)),
                    ToTensor(1),
                    Normalize(get_mean(1), [1, 1, 1])])
					
labels = deepSBD(os.path.join(opt.root_dir, opt.test_subdir, videoname), 32, model, spatial_transforms, 128)

'''




import sys


class STDOUT_TYPE:
    """
    This class represents message types.
    """
    INFO = 1
    ERROR = 2


def printCustom(msg: str, type: int):
    """
    This function represents a customized print function (error/info msg).
    :param msg: Message to print.
    :param type: Type of message (info or error).
    """
    if(type == 1):
        print("INFO: " + msg)
    elif(type == 2):
        print("ERROR: " + msg)
    else:
        print("FATAL ERROR: stdout type does not exist!")
        exit()




def getCommandLineParams():
    """
    This function is used to read commandline parameters (e.g. just used in development stage)
    :return: list of parameters.
    """
    printCustom("read commandline arguments ... ", STDOUT_TYPE.INFO)
    number_of_args = len(sys.argv)
    print('Number of arguments:', len(sys.argv), 'arguments.')
    print('Argument List:', str(sys.argv))

    if (number_of_args < 3):
        printCustom("There must be at least two commandline argument(s)", STDOUT_TYPE.ERROR)
        exit()

    params = sys.argv
    print(params)
    return params




import numpy as np
import cv2
import datetime
from vhh_sbd.utils import *
from PIL import Image


class Video(object):
    """
    This class represents on video and contains properties such as dimensions, length, format, video name ... .
    """

    def __init__(self):
        """
        Constructor.
        """
        #printCustom("create instance of video class ... ", STDOUT_TYPE.INFO);
        self.vidFile = ''
        self.vidName = ""
        self.frame_rate = 0
        self.channels = 0
        self.height = 0
        self.width = 0
        self.format = ''
        self.length = 0
        self.number_of_frames = 0
        self.vid = None
        self.convert_to_gray = False
        self.convert_to_hsv = False

    def load(self, vidFile: str):
        """
        This method is used to load a video of a specified storage.

        :param vidFile: This parameter must hold a valid video file path.
        """
        #printCustom("load video information ... ", STDOUT_TYPE.INFO);
        self.vidFile = vidFile
        if(self.vidFile == ""):
            printCustom("ERROR: you must add a video file path!", STDOUT_TYPE.ERROR)
            exit(1)
        self.vidName = self.vidFile.split('/')[-1]
        self.vid = cv2.VideoCapture(self.vidFile)

        if(self.vid.isOpened() == False):
            printCustom("ERROR: not able to open video file!", STDOUT_TYPE.ERROR)
            exit(1)

        status, frm = self.vid.read()

        self.channels = frm.shape[2]
        self.height = self.vid.get(cv2.CAP_PROP_FRAME_HEIGHT)
        self.width = self.vid.get(cv2.CAP_PROP_FRAME_WIDTH)
        self.frame_rate = self.vid.get(cv2.CAP_PROP_FPS)
        self.format = self.vid.get(cv2.CAP_PROP_FORMAT)
        self.number_of_frames = self.vid.get(cv2.CAP_PROP_FRAME_COUNT)

        self.vid.release()

    def printVIDInfo(self):
        """
        This method is used to print all video properties.
        """
        print("---------------------------------")
        print("Video information")
        print("filename: " + str(self.vidFile))
        print("format: " + str(self.format))
        print("fps: " + str(self.frame_rate))
        print("channels: " + str(self.channels))
        print("width: " + str(self.width))
        print("height: " + str(self.height))
        print("nFrames: " + str(self.number_of_frames))
        print("---------------------------------")

    def getFrame(self, frame_id: int) -> np.ndarray:
        """
        This method is used to return one frame specified with a given frame index.

        :param frame_id: This parameter must hold a valid integer frame index.
        :return: This method returns a numpy frame with a specified index (position).
        """
        self.vid.open(self.vidFile)
        if(frame_id >= self.number_of_frames):
            print("ERROR: frame idx out of range!")
            return []

        #print("Read frame with id: " + str(frame_id));
        time_stamp_start = datetime.datetime.now().timestamp()

        self.vid.set(cv2.CAP_PROP_POS_FRAMES, frame_id)
        status, frame_np = self.vid.read()
        self.vid.release()

        if(status == True):
            if(self.convert_to_gray == True):
                frame_np = cv2.cvtColor(frame_np, cv2.COLOR_BGR2GRAY)
                #print(frame_gray_np.shape);
            if (self.convert_to_hsv == True):
                frame_np = cv2.cvtColor(frame_np, cv2.COLOR_BGR2HSV);
                h, s, v = cv2.split(frame_np)

        time_stamp_end = datetime.datetime.now().timestamp()
        time_diff = time_stamp_end - time_stamp_start
        #print("time: " + str(round(time_diff, 4)) + " sec");

        return frame_np

    def getNumberOfBatches(self) -> int:
        batch_size = 64;
        nBatches = int(self.number_of_frames / batch_size) + 1
        return nBatches;

    def getFrameBatch(self, batch_id: int) -> np.ndarray:
        self.vid.open(self.vidFile);
        if(batch_id >= self.getNumberOfBatches()):
            print("ERROR: batch idx out of range!");
            return [];

        batch_np = []

        for j in range(0, 64):
            frm_id = batch_id * 64 + j;
            frame_np = self.getFrame(frm_id)
            if(len(frame_np) == 0):
                break;
            batch_np.append(frame_np)

        batch_np = np.array(batch_np)
        return batch_np;
      #data: np.ndarray, labels: np.ndarray) -> float:

import torch
from torch.utils import data

class VideoDataset(data.Dataset):

    def __init__(self, vidFile: str, transform=None):
        # printCustom("create instance of video class ... ", STDOUT_TYPE.INFO);
        self.vidFile = '';
        self.vidName = "";
        self.frame_rate = 0;
        self.channels = 0;
        self.height = 0;
        self.width = 0;
        self.format = '';
        self.length = 0;
        self.number_of_frames = 0;
        self.vid = None;
        self.convert_to_gray = False;
        self.convert_to_hsv = False;

        self.transform = transform;

        # printCustom("load video information ... ", STDOUT_TYPE.INFO);
        self.vidFile = vidFile;
        if (self.vidFile == ""):
            printCustom("ERROR: you must add a video file path!", STDOUT_TYPE.ERROR);
            exit(1);
        self.vidName = self.vidFile.split('/')[-1]
        self.vid = cv2.VideoCapture(self.vidFile);

        if (self.vid.isOpened() == False):
            printCustom("ERROR: not able to open video file!", STDOUT_TYPE.ERROR);
            exit(1);

        status, frm = self.vid.read();

        self.channels = frm.shape[2];
        self.height = self.vid.get(cv2.CAP_PROP_FRAME_HEIGHT);
        self.width = self.vid.get(cv2.CAP_PROP_FRAME_WIDTH);
        self.frame_rate = self.vid.get(cv2.CAP_PROP_FPS);
        self.format = self.vid.get(cv2.CAP_PROP_FORMAT);
        self.number_of_frames = self.vid.get(cv2.CAP_PROP_FRAME_COUNT);

        self.vid.release();

    def __len__(self):
        return int(self.number_of_frames)

    def __getitem__(self, frame_id: int):

        self.vid.open(self.vidFile);
        if(frame_id >= self.number_of_frames):
            print("ERROR: frame idx out of range!");
            return [];

        #print("Read frame with id: " + str(frame_id));
        time_stamp_start = datetime.datetime.now().timestamp();

        self.vid.set(cv2.CAP_PROP_POS_FRAMES, frame_id);
        status, frame_np = self.vid.read();
        self.vid.release();

        frame_np = Image.fromarray(frame_np.astype('uint8'))

        time_stamp_end = datetime.datetime.now().timestamp();
        time_diff = time_stamp_end - time_stamp_start;
        #print("time: " + str(round(time_diff, 4)) + " sec");

        if(self.transform != None):
            frame_np = self.transform(frame_np);

        return frame_np;


from vhh_sbd.Video import Video, VideoDataset
from vhh_sbd.utils import *
from vhh_sbd.Configuration import Configuration
from vhh_sbd.Shot import Shot
from vhh_sbd.PreProcessing import PreProcessing
from vhh_sbd.Model import *
from matplotlib import pyplot as plt
from scipy.spatial import distance
from vhh_sbd.DeepSBD import CandidateSelection
import os
import cv2


class SBD(object):
    """
    Main class of shot boundary detection (sbd) package.
    """

    def __init__(self, config_file: str):
        """
        Constructor

        :param config_file: [required] path to configuration file (e.g. PATH_TO/config.yaml)
                                       must be with extension ".yaml"
        """
        #printCustom("INFO: create instance of sbd ... ", STDOUT_TYPE.INFO);

        if (config_file == ""):
            printCustom("No configuration file specified!", STDOUT_TYPE.ERROR)
            exit()

        self.config_instance = Configuration(config_file)
        self.config_instance.loadConfig()

        self.vid_instance = None
        self.pre_proc_instance = PreProcessing(self.config_instance)
        self.candidate_selection_instance = CandidateSelection(self.config_instance)
        self.net = None

        self.src_path = ""
        self.filename_l = ""

    def runOnSingleVideo(self, video_filename, max_recall_id=-1):
        """
        Method to run sbd on specified video.

        :param video_filename: This parameter must hold a valid video file path.
        :param max_recall_id: [required] integer value holding unique video id from VHH MMSI system
        """

        shot_boundaries_l = []
        self.src_path = self.config_instance.path_videos

        vid_name = video_filename.split('/')[-1]

        printCustom("--------------------------", STDOUT_TYPE.INFO)
        printCustom("Process video: " + str(vid_name) + " ... ", STDOUT_TYPE.INFO)

        # load video
        self.vid_instance = Video()
        self.vid_instance.load(video_filename)

        shot_boundaries_np = None
        if (self.config_instance.activate_candidate_selection == 1):
            # candidate selection
            printCustom("Process candidate selection: " + str(vid_name) + " ... ", STDOUT_TYPE.INFO)
            candidate_selection_result_np = self.candidate_selection_instance.run(video_filename)

            # shot boundary detection
            printCustom("Process shot boundary detection: " + str(vid_name) + " ... ", STDOUT_TYPE.INFO)
            shots_np = self.runWithCandidateSelection(candidate_selection_result_np)
            shot_boundaries_l.append(shots_np)
            shot_boundaries_np = np.squeeze(np.array(shot_boundaries_l))

        elif (self.config_instance.activate_candidate_selection == 0):
            # shot boundary detection
            printCustom("Process shot boundary detection: " + str(vid_name) + " ... ", STDOUT_TYPE.INFO)
            shot_boundaries_np = self.runWithoutCandidateSelection(self.src_path, vid_name)


        # convert shot boundaries to final shots
        final_shot_l = self.convertShotBoundaries2Shots(shot_boundaries_np)

        # export final results
        if (self.config_instance.save_final_results == 1):
            self.exportFinalResultsToCsv(final_shot_l, str(max_recall_id))

        return final_shot_l

    def runWithoutCandidateSelection(self, src_path, vid_name):
        """
        This method is used to run sbd without candidate selection mode.

        :param src_path: THis parameter must hold a valid path to the video file.
        :param vid_name: This parameter must hold a valid videofile name.
        :return: This method returns a numpy list with all detected shots in a video.
        """
        #printCustom("process shot detection ... ", STDOUT_TYPE.INFO);

        # load video
        trans = transforms.Compose([transforms.CenterCrop(self.config_instance.resize_dim[0]),
                                    transforms.ToTensor()
                                    ])
        self.vid_instance = VideoDataset(src_path + "/" + vid_name, transform=trans)

        # initial pre-trained model
        self.net = PyTorchModel(model_arch=self.config_instance.backbone_cnn, use_gpu=self.config_instance.use_gpu)

        # read all frames of video
        cap = cv2.VideoCapture(src_path + "/" + vid_name)
        frame_l = []

        cnt = 0
        while(True):
            cnt = cnt + 1
            ret, frame = cap.read()
            #print(cnt)
            #print(ret)
            #print(cap.get(cv2.CAP_PROP_FRAME_COUNT))
            if (ret == True):
                frame = self.pre_proc_instance.applyTransformOnImg(frame)
                frame_l.append(frame)

                if (self.config_instance.debug_flag == 1):
                    if(cnt <= 500):
                        cv2.imwrite(self.config_instance.path_eval_results + vid_name + "_" + str(cnt) + ".png", frame)

            else:
                break
        #exit()
        frame_np = np.array(frame_l)

        # calculate features and similarities
        number_of_frames = len(frame_np)

        #number_of_frames = 1000
        results_l = []
        shot_l = []
        for i in range(1, number_of_frames):
            #print("-------------------")
            #print("process " + str(i))
            idx_curr = i
            idx_prev = i - 1

            frm_prev = frame_np[idx_prev]
            frm_curr = frame_np[idx_curr]

            # print("process core part ... ")
            feature_prev = self.net.getFeatures(frm_prev)
            feature_curr = self.net.getFeatures(frm_curr)
            result = self.calculateDistance(feature_prev, feature_curr)
            # print(result)

            if (int(self.config_instance.save_raw_results) == 1):
                results_l.append(result)

        # calculate thresholds
        distances_np = np.array(results_l)
        if (self.config_instance.threshold_mode == 'adaptive'):
            thresholds = []
            window_size = self.config_instance.window_size
            alpha = self.config_instance.threshold
            for i in range(0, len(distances_np)):

                if(i % window_size == 0):
                    print(i)
                    #print(distances_np)
                    print(distances_np[i:i+window_size])
                    th = np.mean(distances_np[i:i+window_size]) * 6.0
                    print(th)

                thresholds.append(th)
            thresholds = np.array(thresholds)
            print(thresholds.shape)
            #exit()
            for i in range(0, len(distances_np)):
                #print("####################")
                #print(i)
                #print("th: " + str(thresholds[i]))
                #print("dist: " + str(distances_np[i]))

                if (distances_np[i] > thresholds[i]):
                    idx_curr = i + 1
                    idx_prev = i

                    print("cut at: " + str(i) + " -> " + str(i+1))

                    printCustom("Abrupt Cut detected: " + str(idx_prev) + ", " + str(idx_curr), STDOUT_TYPE.INFO)
                    shot_l.append([self.vid_instance.vidName, (idx_prev, idx_curr)])
        elif (self.config_instance.threshold_mode == 'fixed'):
            for i in range(0, len(distances_np)):
                if (distances_np[i] > self.config_instance.threshold):
                    idx_curr = i + 1
                    idx_prev = i

                    # print("cut at: " + str(i) + " -> " + str(i+1))
                    # print(i)
                    # print(thresholds[i])
                    # print(distances_np[i])
                    printCustom("Abrupt Cut detected: " + str(idx_prev) + ", " + str(idx_curr), STDOUT_TYPE.INFO)
                    shot_l.append([self.vid_instance.vidName, (idx_prev, idx_curr)])

        # save raw results to file
        if (int(self.config_instance.save_raw_results) == 1):
            print("save raw results ... ")
            raw_results_l = []
            raw_results_l.append([1, number_of_frames, results_l])
            results_np = np.array(raw_results_l)
            if (self.config_instance.path_postfix_raw_results == 'csv'):
                self.exportRawResultsAsCsv_New(results_np)
            elif (self.config_instance.path_postfix_raw_results == 'npy'):
                self.exportRawResultsAsNumpy(results_np)

        if (len(shot_l) == 0):
            print("no cuts detected ... ")
            shot_l.append([self.vid_instance.vidName, (-1, -1)])

        # convert shot boundaries to shots
        shots_np = np.array(shot_l)

        return shots_np

    def runWithCandidateSelection(self, candidates_np):
        """
        This method is used to run sbd with candidate selection mode.

        :param candidates_np: THis parameter must hold a valid numpy list including all pre-selected candidates.
        :return: This method returns a numpy list with all detected shots in a video.
        """
        #printCustom("process shot detection ... ", STDOUT_TYPE.INFO);

        # initial pre-trained model
        self.net = PyTorchModel(model_arch=self.config_instance.backbone_cnn, use_gpu=self.config_instance.use_gpu)

        results_l = []
        shot_l = []
        for i in range(0, len(candidates_np)):
            start = candidates_np[i][0]
            end = candidates_np[i][1] - 1

            #print(start)
            #print(end)
            results_per_range = []
            for j in range(start+1, end):
                # print("-------------------")
                # print("process " + str(i))
                idx_curr = j
                idx_prev = j - 1

                frm_prev = self.vid_instance.getFrame(idx_prev)
                frm_curr = self.vid_instance.getFrame(idx_curr)

                if(len(frm_prev) == 0 or len(frm_curr) == 0):
                    break
                #print(idx_prev)
                #print(idx_curr)
                #print(frm_prev.shape)
                #print(frm_curr.shape)

                frm_trans_prev = self.pre_proc_instance.applyTransformOnImg(frm_prev)
                frm_trans_curr = self.pre_proc_instance.applyTransformOnImg(frm_curr)

                # print("process core part ... ")
                feature_prev = self.net.getFeatures(frm_trans_prev)
                feature_curr = self.net.getFeatures(frm_trans_curr)
                result = self.calculateDistance(feature_prev, feature_curr)

                if (int(self.config_instance.save_raw_results) == 1):
                    results_per_range.append(result)

                if(result >= self.config_instance.threshold):
                    printCustom("Abrupt Cut detected: " + str(idx_prev) + ", " + str(idx_curr), STDOUT_TYPE.INFO)
                    shot_l.append([self.vid_instance.vidName, (idx_prev, idx_curr)])

            if (int(self.config_instance.save_raw_results) == 1):
                results_l.append([start, end, results_per_range])

        # save raw results to file
        if (int(self.config_instance.save_raw_results) == 1):
            print("save raw results ... ")
            results_np = np.array(results_l)
            self.exportRawResultsAsCsv_New(results_np)

        # convert shot boundaries to shots
        shots_np = np.array(shot_l)

        return shots_np

    def exportFinalResultsToCsv(self, shot_l: list, name: str):
        """
        This method is used to export the final results to a csv file (semicolon seperated).
        :param shot_l: This parameter must hold a valid array list including the final results list.
        :param name: This parameter represents the name of the csv list.
        """
        printCustom("Export shot list to csv file ... ", STDOUT_TYPE.INFO)

        fp = open(self.config_instance.path_final_results + "/" + str(name) + ".csv", 'w')   # final_shots_"
        fp.write("vid_name;shot_id;start;end" + "\n")
        for shot in shot_l:
            tmp_str = shot.convert2String()
            fp.write(tmp_str + "\n")
        fp.close()

    def exportShotBoundariesResultsToCsv(self, shot_boundaries_np: np.ndarray, name: str):
        """
        This method is used to export the final results to a csv file (semicolon seperated).
        :param shot_l: This parameter must hold a valid array list including the final results list.
        :param name: This parameter represents the name of the csv list.
        """
        printCustom("Export shot list to csv file ... ", STDOUT_TYPE.INFO)

        fp = open("/data/share/maxrecall_vhh_mmsi/develop/videos/results/sbd/" + "/" + str(name) + "_sbd.csv", 'w')  # final_shots_"
        fp.write("shot_id;vid_name;start;end" + "\n")
        for i in range(0, len(shot_boundaries_np)):
            video_name = shot_boundaries_np[i][0]
            start, end = shot_boundaries_np[i][1]
            tmp_str = str(video_name) + ";" + str(start) + ";" + str(end)
            fp.write(tmp_str + "\n")
        fp.close()

    def exportRawResultsAsNumpy(self, results_np: np.ndarray):
        """
        This method is used to export the raw results to a numpy file.

        :param results_np: This parameter must hold a valid numpy list including the raw results.
        """
        np.save(self.config_instance.path_raw_results +
                  self.config_instance.path_prefix_raw_results +
                  str(self.vid_instance.vidName.split('.')[0]) + "." +
                  self.config_instance.path_postfix_raw_results,
                results_np)

    def exportRawResultsAsCsv_New(self, results_np: np.ndarray):
        """
        This method is used to export the raw results to a csv file.

        :param results_np: This parameter must hold a valid numpy list including the raw results.
        """
        # save raw results to file
        fp = open(self.config_instance.path_raw_results +
                  self.config_instance.path_prefix_raw_results +
                  str(self.vid_instance.vidName.split('.')[0]) + "." +
                  self.config_instance.path_postfix_raw_results, mode='w')

        for i in range(0, len(results_np)):
            start, end, distances_l = results_np[i]
            tmp_str = str(start) + ";" + str(end)
            for j in range(0, len(distances_l)):
                tmp_str = tmp_str + ";" + str(distances_l[j])
            fp.write(self.vid_instance.vidName.split('.')[0] + ";" + str(tmp_str) + "\n")
            # csv_writer.writerow(row);
        fp.close()

    def calculateDistance(self, x, y):
        """
        This method is used to calculate the distance between 2 feature vectors.

        :param x: This parameter represents a feature vector (one-dimensional)
        :param y: This parameter represents a feature vector (one-dimensional)
        :return: This method returns the similarity score of a specified distance metric.
        """
        dst = 0

        # initial pre-trained model
        if (self.config_instance.similarity_metric == "cosine"):
            dst = distance.cosine(x, y)
        elif (self.config_instance.similarity_metric == "euclidean"):
            dst = distance.euclidean(x, y)
        else:
            dst = None
            printCustom("No valid similarity metric selected!", STDOUT_TYPE.ERROR)
            exit()

        return dst

    def convertShotBoundaries2Shots(self, shot_boundaries_np: np.ndarray):
        """
        This method converts a list with detected shot boundaries to the final shots.

        :param shot_boundaries_np: This parameter must hold a numpy array with all detected shot boundaries.
        :return: This method returns a numpy list with the final shots.
        """
        # convert results to shot instances

        shot_l = []

        vidname_curr = shot_boundaries_np[0][0]
        start_curr, stop_curr = shot_boundaries_np[0][1]

        print("shot boundaries list: ")
        print(shot_boundaries_np)

        if (start_curr == -1 and stop_curr == -1):
            print("no shots detected ... ")
            shot = Shot(len(shot_boundaries_np), vidname_curr, 0, int(self.vid_instance.number_of_frames) - 1)
            shot_l.append(shot)
            return shot_l

        shot_start = 0
        shot_end = start_curr
        shot = Shot(1, vidname_curr, shot_start, shot_end)
        shot_l.append(shot)

        for i in range(1, len(shot_boundaries_np)):
            #print(i)
            start_prev, stop_prev = shot_boundaries_np[i-1][1]
            start_curr, stop_curr = shot_boundaries_np[i][1]
            vidname_curr = shot_boundaries_np[i][0]

            shot_start = int(stop_prev)
            shot_end = int(start_curr)
            shot = Shot(i + 1, vidname_curr, shot_start, shot_end)
            shot_l.append(shot)

        vidname_curr = shot_boundaries_np[-1][0]
        start_curr, stop_curr = shot_boundaries_np[-1][1]
        shot_start = int(stop_curr)
        shot_end = int(self.vid_instance.number_of_frames) - 1
        shot = Shot(len(shot_boundaries_np) + 1, vidname_curr, shot_start, shot_end)
        shot_l.append(shot)

        if(self.config_instance.debug_flag == 1):
            # print shot infos
            for i in range(0, len(shot_l)):
                shot_l[i].printShotInfo()

        return shot_l
